import PyAsterisk

class PyAsteriskXMLRPC:
	"""
	$Id: PyAsteriskXMLRPC.py,v 1.2 2004/10/06 05:33:33 ww Exp $

	class PyAsteriskXMLRPC
	
	This exposes functionality of and information about the Asterisk PBX. It
	is a partial abstraction of the PyAsterisk class providing a subset of
	the available functions using XMLRPC. It is intended to be used with the
	SimpleXMLRPCServer or SecureXMLRPCServer classes.

	This file is part of the PyAsterisk distribution which is available at
	
	        http://vox.groovy.net/moin/PyAsterisk

	This program is Free Software, provided WITHOUT WARRANTY and released
	subject to the terms of the GNU GPL published at
	
	        http://www.gnu.org/licenses/gpl.html
		
	and as may be amended from time to time.

	Copyleft (.) 2004 Consultants Ars Informatica S.A.R.F.
	"""
	def __init__(self, server):
		self.__server__ = server
		
	_loglevels = ["LOG_DEBUG", "LOG_EVENT", "LOG_NOTICE", "LOG_WARNING", "LOG_ERROR"]
	def log(self, level, message):
		"""
		Send a log message. level must be a string that is contained
		in the _loglevels class variable.
		"""
		if level not in _loglevels:
		    raise TypeError, "level must be one of %s" % (loglevels,)
		level = eval(level)
		PyAsterisk.log(level, message)
		return 0
	def verbose(self, message):
		"""
		Send debugging or informational message to the asterisk
		console.
		"""
		PyAsterisk.verbose(message)
		return 0

	###
	### channel operations
	###
	def active_channels(self):
		"""
		return the number of currently active channels.
		equivalent to len(channels()) but more efficient.
		"""
		return PyAsterisk.active_channels()
	
	def channels(self):
		"""
		Return a list of currently existing channels.
	        """
		return PyAsterisk.channels()

	def hangup(self, channel):
		"""
		Try to hang up the named channel. Raises a KeyError if
		the channel does not exits. Returns 0 on success, -1 on
		failure.
		"""
		chan = PyAsterisk.PyAsteriskChannel(channel)
		chan.hangup()
		return 0

	def getvar(self, channel, var):
		"""
		Return a the value of a channel variable. Raises a KeyError if
		the channel does not exist, returns the integer 0 if the variable
		is unset, returns the string representation of the variable if it
		is set.
		"""
		chan = PyAsterisk.PyAsteriskChannel(channel)
		val = chan[var]
		if not val:
			return 0
		return val
	
	def setvar(self, channel, var, value):
		"""
		Set the value of a channel variable. Raises a KeyError if the
		channel does not exist. Value will be coerced to a string. Returns
		the integer 0.
		"""
		chan = PyAsterisk.PyAsteriskChannel(channel)
		chan[var] = str(value)
		return 0

	def getcid(self, channel):
		"""
		Get the CallerID value for the named channel.
		"""
		chan = PyAsterisk.PyAsteriskChannel(channel)
		return chan.getcid()
	def getseconds(self, channel):
		"""
		Get the number of seconds the call has been active.
		"""
		chan = PyAsterisk.PyAsteriskChannel(channel)
		return chan.getseconds()

	###
	### dialplan operations
	###
	def contexts(self):
		"""
		Return a list of configured contexts in the dial plan.
		"""
		return PyAsterisk.contexts()
	def includes(self, context):
		"""
		Return a list of includes in the given dialplan context.
		"""
		ctx = PyAsterisk.PyAsteriskContext(context)
		return ctx.includes()
	def include_add(self, context, include):
		"""
		Adds the include to the given context.
		"""
		ctx = PyAsterisk.PyAsteriskContext(context)
		return ctx.include_add(include)
	def include_del(self, context, include):
		"""
		Deletes the include from the given context.
		"""
		ctx = PyAsterisk.PyAsteriskContext(context)
		return ctx.include_del(include)
	def extensions(self, context):
		"""
		Returns a list (exten, cid) of extensions in the given
		context.
		"""
		ctx = PyAsterisk.PyAsteriskContext(context)
		return ctx.extensions()
	def extension_get(self, context, exten):
		"""
		Returns a list, [[prio, app, data], ...]
		For the given extension.
		"""
		ctx = PyAsterisk.PyAsteriskContext(context)
		return ctx.extension_get(exten)
	def extension_add(self, context, exten, prio, app, data=""):
		"""
		Add the given extension to the context.
		"""
		ctx = PyAsterisk.PyAsteriskContext(context)
		return ctx.extension_add(exten, prio, app, data)

	def extension_del(self, context, exten, prio):
		"""
		Delete the given extension from the context.
		"""
		ctx = PyAsterisk.PyAsteriskContext(context)
		return ctx.extension_del(exten, prio)
	
	def redirect(self, channel, cep):
		"""
		Redirect the channel to the given (context, extension, priority) triple.
		Raises a KeyError if the channel does not exist. context and extension are
		strings, priority is an integer. Returns the -1 on failure, 0 on success.
		"""
		chan = PyAsterisk.PyAsteriskChannel(channel)
		return chan.redirect(cep)
	
	def originate(self, channel, dest, timeout=None, callerid=None, variables=None, account=None):
		"""
		Originate a call using the given channel to the given destination.

		channel is of the form (medium, params) where medium is in ['SIP', 'IAX2', 'OH323', 'Zap', ...]
		and params is an appropriate string that will be accepted by the Asterisk C API.

		dest is either of the form (context, extension, priority) or (app, data) according to
		whether you want the call to be connected directly to a particular application (such
		as a MeetMe conference or hold queue, or else if the call should be routed according to
		the normal extension logic.
		
		Optional parameters:
		
		timeout is in milliseconds and defaults to 30000.
		callerid is a string of the form "Your Name <12105551212>"
		variables is a dictionary of channel variables to set.
		account is the account code to use with the default set at compile time.
		"""
		from string import join

		kwargs = {}
		if variables:
			variables = join(map(lambda x: "%s=%s" % (x, variables[x]), variables.keys()), '|')
			kwargs['variables'] = variables
		if timeout:
			kwargs['timeout'] = timeout
		if callerid:
			kwargs['callerid'] = callerid
		if account:
			kwargs['account'] = account
			
		if len(dest) == 2:
			PyAsterisk.originate_app(*(channel + dest), **kwargs)
		elif len(dest) == 3:
			PyAsterisk.originate_exten(*(channel + dest), **kwargs)
		else:
			raise TypeError, "destination must be (app, appdata) or (context, exten, prio)"
		return 0
