/*
 * app_python.c -- Embedded Python Interpreter for Asterisk
 *
 * This software is distributed under the terms of the GNU GPL, version 2.0 or
 * later, a copy of which you should have recieved in the source distribution 
 * with this file or which is also available at http://www.fsf.org/licences/gpl.html,
 * with one additional clause:
 *
 * Permission is granted to link this program at runtime with the chan_g729a.so
 * library distributed by Digium Inc. in ftp://ftp.digium.com/pub/asterisk/g729
 * or any other implementation of the proprietary and patent-encumbered pseudo
 * standard known as G.729, G.729A or G.729B.
 *
 */

#include <Python.h>
#include <pthread.h>

#include <asterisk/pbx.h>
#include <asterisk/config.h>
#include <asterisk/channel.h>
#include <asterisk/module.h>
#include <asterisk/logger.h>
#include <asterisk/lock.h>

#include "py_asterisk.h"

#define PyASTERISK_VERSION "$Id: app_python.c,v 1.1 2004/10/06 04:47:17 ww Exp $"

#ifdef ASTERISK_1_0
#define ast_pthread_create pthread_create
#endif

static char tdesc[] = "A Python in the DialPlan";
static char app[] = "Python";
static char synopsis[] = "Execute a python script";
static char descrip[] =
"  Python(foo[, arg1, arg2, ...])\n\n"
"This will cause the embedded python interpreter to import the\n"
"execute() method from the 'foo' module which must be in sys.path\n"
"The method is called as foo.execute(chan, args...) where args are\n"
"optionally passed in the dialplan as above, and chan is the name\n"
"of the channel that can be used as a key in the PyAsterisk API\n";

static int app_python(struct ast_channel *chan, void *data)
{
  PyInterpreterState *interp;
  PyThreadState *tstate;
  char buf[BUFSIZ];

  snprintf(buf, sizeof(buf),
	   "try:\n"
	   "    from %s import execute\n"
	   "    execute('%s')\n"
	   "except:\n"
	   "    from traceback import format_exception\n"
	   "    from sys import exc_info\n"
	   "    from string import join\n"
	   "    from _asterisk import verbose\n"
	   "    verbose(join(apply(format_exception, exc_info()), '\\n'))\n",
	   (char *)data, chan->name);

  interp = PyAsteriskInterpreter();
  if(!interp)
    return -1;

  tstate = PyThreadState_New(interp);
  if(!tstate)
    return -1;
  PyEval_AcquireThread(tstate);

  PyRun_SimpleString(buf);

  PyEval_ReleaseThread(tstate);
  PyThreadState_Delete(tstate);

  return 0;
}

/*
 * Asterisk Module Housekeeping. Ver' simple.
 */
  
STANDARD_LOCAL_USER;
LOCAL_USER_DECL;

int load_module(void)
{
  int res;
  res = ast_register_application(app, app_python, synopsis, descrip);
  return 0;
}

int unload_module(void)
{
  STANDARD_HANGUP_LOCALUSERS;
  return ast_unregister_application(app);
}

char *description(void)
{
  return tdesc;
}

int usecount(void)
{
  int res;
  STANDARD_USECOUNT(res);
  return res;
}

char *key()
{
  return ASTERISK_GPL_KEY;
}
