/*
 * res_python.c -- Embedded Python Interpreter for Asterisk
 *
 * This software is distributed under the terms of the GNU GPL, version 2.0 or
 * later, a copy of which you should have recieved in the source distribution 
 * with this file or which is also available at http://www.fsf.org/licences/gpl.html,
 * with one additional clause:
 *
 * Permission is granted to link this program at runtime with the chan_g729a.so
 * library distributed by Digium Inc. in ftp://ftp.digium.com/pub/asterisk/g729
 * or any other implementation of the proprietary and patent-encumbered pseudo
 * standard known as G.729, G.729A or G.729B.
 *
 */

#include <Python.h>
#include <pthread.h>

#include <asterisk/pbx.h>
#include <asterisk/config.h>
#include <asterisk/channel.h>
#include <asterisk/module.h>
#include <asterisk/logger.h>
#include <asterisk/lock.h>

#include "py_asterisk.h"

#define PyASTERISK_VERSION "$Id: res_python.c,v 1.5 2004/10/06 05:33:22 ww Exp $"

#ifdef ASTERISK_1_0
#define ast_pthread_create pthread_create
#endif

static char tdesc[] = "Asterisk Embedded Python Interpreter";
static pthread_t interp_thread;
static PyInterpreterState *python_interp = NULL;

PyInterpreterState *PyAsteriskInterpreter(void)
{
  return python_interp;
}

static PyObject *PyAsteriskLog(PyObject *self, PyObject *args)
{
  int level;
  const char *s;
  if(!PyArg_ParseTuple(args, "is", &level, &s))
    return NULL;
  ast_log(level, _A_, "%s\n", s);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *PyAsteriskVerbose(PyObject *self, PyObject *args)
{
	const char *s;
	if(!PyArg_ParseTuple(args, "s", &s))
		return NULL;
	ast_verbose("%s\n", s);
	Py_INCREF(Py_None);	
	return Py_None;
}

static PyObject *PyAsteriskActiveChannels(PyObject *self, PyObject *args)
{
  if(!PyArg_ParseTuple(args, ""))
    return NULL;
  return Py_BuildValue("i", ast_active_channels());
}

static PyObject *PyAsteriskChannels(PyObject *self, PyObject *args)
{
  PyObject *ret;
  PyObject *o;
  struct ast_channel *c = NULL;

  if(!PyArg_ParseTuple(args, ""))
    return NULL;

  ret = PyList_New(0);
  if(!ret) {
    PyErr_SetString(PyExc_RuntimeError, "Out of memory");
    return NULL;
  }

  c = ast_channel_walk_locked(NULL);
  while(c) {
    o = Py_BuildValue("s", c->name);
    ast_mutex_unlock(&c->lock);
    if(!o) {
      PyErr_SetString(PyExc_RuntimeError, "Out of memory");
      return NULL;
    }
    if(PyList_Append(ret, o))
      return NULL;
    c = ast_channel_walk_locked(c);
  }
  return ret;
}

static PyObject *PyAsteriskChannelHangup(PyObject *self, PyObject *args)
{
  char *name;
  struct ast_channel *c = NULL;

  if(!PyArg_ParseTuple(args, "s", &name))
    return NULL;
  c = ast_get_channel_by_name_locked(name);
  if(!c) {
    PyErr_SetString(PyExc_KeyError, "No such channel");
    return NULL;
  }
  ast_softhangup(c, AST_SOFTHANGUP_EXPLICIT);
  ast_mutex_unlock(&c->lock);
  
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *PyAsteriskChannelSetVar(PyObject *self, PyObject *args)
{
  char *name;
  char *varname;
  char *varval;
  struct ast_channel *c;

  if(!PyArg_ParseTuple(args, "sss", &name, &varname, &varval))
    return NULL;
  c = ast_get_channel_by_name_locked(name);
  if(!c) {
    PyErr_SetString(PyExc_KeyError, "No such channel");
    return NULL;
  }
  pbx_builtin_setvar_helper(c,varname,varval);
  ast_mutex_unlock(&c->lock);
  
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *PyAsteriskChannelGetVar(PyObject *self, PyObject *args)
{
  char *name;
  char *varname;
  char *varval;
  struct ast_channel *c;

  if(!PyArg_ParseTuple(args, "ss", &name, &varname))
    return NULL;
  c = ast_get_channel_by_name_locked(name);
  if(!c) {
    PyErr_SetString(PyExc_KeyError, "No such channel");
    return NULL;
  }
  varval = pbx_builtin_getvar_helper(c,varname);
  ast_mutex_unlock(&c->lock);

  return Py_BuildValue("s", varval);
}

static PyObject *PyAsteriskChannelSetAttr(PyObject *self, PyObject *args)
{
  char *name;
  char *attrname;
  char *attrval;
  struct ast_channel *c;

  if(!PyArg_ParseTuple(args, "sss", &name, &attrname, &attrval))
    return NULL;
  c = ast_get_channel_by_name_locked(name);
  if(!c) {
    PyErr_SetString(PyExc_KeyError, "No such channel");
    return NULL;
  }

  ast_mutex_unlock(&c->lock);
  
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *PyAsteriskChannelGetAttr(PyObject *self, PyObject *args)
{
  char *name;
  char *attrname;
  struct ast_channel *c;
  PyObject *ret = NULL;
  
  if(!PyArg_ParseTuple(args, "ss", &name, &attrname))
    return NULL;
  c = ast_get_channel_by_name_locked(name);
  if(!c) {
    PyErr_SetString(PyExc_KeyError, "No such channel");
    return NULL;
  }

  if(!strcasecmp(attrname, "bridge")) {
    ret = Py_BuildValue("s", c->bridge->name);
  } else if(!strcasecmp(attrname, "seconds") && c->pbx) {
    struct timeval now;
    gettimeofday(&now, NULL);
    ret = Py_BuildValue("i", now.tv_sec - c->cdr->start.tv_sec);
  } else if(!strcasecmp(attrname, "callerid")) {
    ret = Py_BuildValue("s", c->callerid ? c->callerid : "<unknown>");
  } else if(!strcasecmp(attrname, "state")) {
    ret = Py_BuildValue("s", ast_state2str(c->_state));
  } else if(!strcasecmp(attrname, "account")) {
    ret = Py_BuildValue("s", c->accountcode);
  } else if(!strcasecmp(attrname, "cep") && c->pbx) {
    ret = Py_BuildValue("(ssi)", c->context, c->exten, c->priority);
  } else if(!strcasecmp(attrname, "uniqueid")) {
    ret = Py_BuildValue("s", c->uniqueid); 
  }

  ast_mutex_unlock(&c->lock);

  if(!ret)
    ret = Py_BuildValue("i", 0);

  return ret;
}

static PyObject *PyAsteriskRedirect(PyObject *self, PyObject *args)
{
  char *name, *context, *exten;
  int prio, res;

  struct ast_channel *c;

  if(!PyArg_ParseTuple(args, "s(ssi)", &name, &context, &exten, &prio))
    return NULL;

  c = ast_get_channel_by_name_locked(name);
  if(!c) {
    PyErr_SetString(PyExc_KeyError, "No such channel");
    return NULL;
  }
  res = ast_async_goto(c, context, exten, prio);
  ast_mutex_unlock(&c->lock);

  return Py_BuildValue("i", res);
}

static PyObject *PyAsteriskOriginateApp(PyObject *self, PyObject *args, PyObject *kwords)
{
  int res;
  int timeout = 30000;
  char *tech, *data, *app, *appdata;
  char *callerid = NULL;
  char *variable = "";
  char *account = PyASTERISK_DEFAULT_ACCOUNTCODE;

  int reason;
  static char *kwlist[] = {"tech", "data", "app", "appdata",
			   "timeout", "callerid", "variable", "account", NULL};

  if(!PyArg_ParseTupleAndKeywords(args, kwords, "ssss|isss", kwlist,
				  &tech, &data, &app, &appdata,
				  &timeout, &callerid, &variable, &account))
    return NULL;

  res = ast_pbx_outgoing_app(tech, AST_FORMAT_SLINEAR, data, timeout, app, appdata, 
			     &reason, 1, callerid, variable, account);
  return Py_BuildValue("i", res);
}

static PyObject *PyAsteriskOriginateExten(PyObject *self, PyObject *args, PyObject *kwords)
{
  int res;
  int timeout = 30000;
  char *tech, *data, *context, *exten;
  int prio;
  char *callerid = NULL;
  char *variable = "";
  char *account = PyASTERISK_DEFAULT_ACCOUNTCODE;
  int reason = 0;

  static char *kwlist[] = {"tech", "data", "context", "exten", "prio", 
			   "timeout", "callerid", "variable", "account", NULL};

  if(!PyArg_ParseTupleAndKeywords(args, kwords, "ssssi|isss", kwlist,
				  &tech, &data, &context, &exten, &prio,
				  &timeout, &callerid, &variable, &account))
    return NULL;

  res = ast_pbx_outgoing_exten(tech, AST_FORMAT_SLINEAR, data, timeout, context, exten, prio,
			       &reason, 1, callerid, variable, account);
  return Py_BuildValue("i", res);
}

static PyObject *PyAsteriskContexts(PyObject *self, PyObject *args)
{
  struct ast_context *c;
  PyObject *ret;

  if(!PyArg_ParseTuple(args, ""))
    return NULL;

  ret = PyList_New(0);
  if(!ret)
    return NULL;

  if(ast_lock_contexts()) {
    Py_DECREF(ret);
    PyErr_SetString(PyExc_RuntimeError, "Unable to lock context list");
    return NULL;
  }

  c = ast_walk_contexts(NULL);
  while (c) {
    ast_lock_context(c);
    if(PyList_Append(ret, Py_BuildValue("s", ast_get_context_name(c)))) {
      Py_DECREF(ret);
      ast_unlock_context(c);
      return NULL;
    }
    ast_unlock_context(c);
    c = ast_walk_contexts(c);
  }
  ast_unlock_contexts();

  return ret;
}

static PyObject *PyAsteriskContextIncludes(PyObject *self, PyObject *args)
{
  PyObject *ret;
  char *context;
  struct ast_context *ctx;
  struct ast_include *inc;

  if(!PyArg_ParseTuple(args, "s", &context))
    return NULL;

  ctx = ast_context_find(context);
  if(!ctx) {
    PyErr_SetString(PyExc_KeyError, "No such context");
    return NULL;
  }

  ret = PyList_New(0);
  if(!ret) {
    return NULL;
  }

  ast_lock_context(ctx);
  inc = ast_walk_context_includes(ctx, NULL);
  while(inc) {
    if(PyList_Append(ret, Py_BuildValue("s", ast_get_include_name(inc)))) {
      Py_DECREF(ret);
      ast_unlock_context(ctx);
      return NULL;
    }
    inc = ast_walk_context_includes(ctx, inc);
  }
  ast_unlock_context(ctx);

  return ret;
}

static PyObject *PyAsteriskContextIncludeAdd(PyObject *self, PyObject *args)
{
  char *context, *include, *registrar = PyASTERISK_NAME;

  if(!PyArg_ParseTuple(args, "ss|s", &context, &include, &registrar))
    return NULL;

  return Py_BuildValue("i", ast_context_add_include(context, include, registrar));
}

static PyObject *PyAsteriskContextIncludeDel(PyObject *self, PyObject *args)
{
  char *context, *include, *registrar = PyASTERISK_NAME;
  
  if(!PyArg_ParseTuple(args, "ss|s", &context, &include, &registrar))
    return NULL;

  return Py_BuildValue("i", ast_context_remove_include(context, include, registrar));
}


static PyObject *PyAsteriskContextExtensions(PyObject *self, PyObject *args)
{
  PyObject *ret;
  char *context;
  struct ast_context *ctx;
  struct ast_exten *exten;

  if(!PyArg_ParseTuple(args, "s", &context))
    return NULL;

  ctx = ast_context_find(context);
  if(!ctx) {
    PyErr_SetString(PyExc_KeyError, "No such context");
    return NULL;
  }

  ret = PyList_New(0);
  if(!ret) {
    return NULL;
  }

  ast_lock_context(ctx);
  exten = ast_walk_context_extensions(ctx, NULL);
  while(exten) {
    PyObject *ext;

    ext = Py_BuildValue("s", ast_get_extension_name(exten));
    if(!ext) {
      Py_DECREF(ret);
      ast_unlock_context(ctx);
      return NULL;
    }

    if(PyList_Append(ret, ext)) {
      Py_DECREF(ret);
      ast_unlock_context(ctx);
      return NULL;
    }
    exten = ast_walk_context_extensions(ctx, exten);
  }
  ast_unlock_context(ctx);

  return ret;
}

static PyObject *PyAsteriskContextExtensionGet(PyObject *self, PyObject *args)
{
  char *context, *exten;
  PyObject *ret, *prio, *app, *data, *registrar, *tmp;
#ifndef ASTERISK_1_0
  PyObject *cid;
#endif

  struct ast_context *ctx;
  struct ast_exten *ext, *extp;

  if(!PyArg_ParseTuple(args, "ss", &context, &exten))
    return NULL;
  ctx = ast_context_find(context);
  if(!ctx) {
    PyErr_SetString(PyExc_KeyError, "No such context");
    return NULL;
  }
  
  ret = PyList_New(0);
  if(!ret) {
    return NULL;
  }

  ast_lock_context(ctx);
  ext = ast_walk_context_extensions(ctx, NULL);
  while(ext) {
    if(!strcmp(exten, ast_get_extension_name(ext)))
      break;
    ext = ast_walk_context_extensions(ctx, ext);
  }
  if(!ext) {
    Py_DECREF(ret);
    PyErr_SetString(PyExc_KeyError, "No such extension");
    ast_unlock_context(ctx);
    return NULL;
  }
  
  extp = ast_walk_extension_priorities(ext, NULL);
  while(extp) {
#ifdef ASTERISK_1_0
    tmp = PyTuple_New(4);
#else
    tmp = PyTuple_New(5);
#endif
    if(!tmp) {
      Py_DECREF(ret);
      ast_unlock_context(ctx);
      return NULL;
    }

    prio = Py_BuildValue("i", ast_get_extension_priority(extp));
    if(!prio) goto cleanup;
    PyTuple_SET_ITEM(tmp, 0, prio);

    app = Py_BuildValue("s", ast_get_extension_app(extp));
    if(!app) goto cleanup;
    PyTuple_SET_ITEM(tmp, 1, app);

    data = Py_BuildValue("s", ast_get_extension_app_data(extp));
    if(!data) goto cleanup;
    PyTuple_SET_ITEM(tmp, 2, data);

    registrar = Py_BuildValue("s", ast_get_extension_registrar(extp));
    if(!registrar) goto cleanup;
    PyTuple_SET_ITEM(tmp, 3, registrar);

#ifndef ASTERISK_1_0
    cid = Py_BuildValue("s", ast_get_extension_cidmatch(extp));
    if(!cid) goto cleanup;
    PyTuple_SET_ITEM(tmp, 4, data);
#endif

    if(PyList_Append(ret, tmp)) goto cleanup;

    extp = ast_walk_extension_priorities(ext, extp);

    continue;

  cleanup:
    Py_DECREF(tmp);
    Py_DECREF(ret);
    ast_unlock_context(ctx);
    return NULL;
  }

  ast_unlock_context(ctx);
  return ret;
}

static PyObject *PyAsteriskContextExtensionAdd(PyObject *self, PyObject *args)
{
  char *context, *exten, *app, *data = "", *cid = NULL, *registrar = PyASTERISK_NAME;
  int prio, replace = 1;
  int ret;

  if(!PyArg_ParseTuple(args, "ssis|siss", 
		       &context, &exten, &prio, &app, &data, &replace, &cid, &registrar))
    return NULL;

  ret = ast_add_extension(context, replace, exten, prio, cid, app, (void *)strdup(data), free, registrar);
  return Py_BuildValue("i", ret);
}

static PyObject *PyAsteriskContextExtensionDel(PyObject *self, PyObject *args)
{
  char *context, *exten, *registrar = PyASTERISK_NAME;
  int prio, ret;

  if(!PyArg_ParseTuple(args, "ssi|s", &context, &exten, &prio, &registrar))
    return NULL;
  ret = ast_context_remove_extension(context, exten, prio, registrar);
  return Py_BuildValue("i", ret);
}

static PyMethodDef AsteriskMethods [] = {
  {"log", PyAsteriskLog, METH_VARARGS,
   "log(LEVEL, string) --> send a log message to the Asterisk logger."},
  {"verbose", PyAsteriskVerbose, METH_VARARGS,
   "verbose(string) --> send a verbose informational message to the Asterisk console."},
  {"active_channels", PyAsteriskActiveChannels, METH_VARARGS,
   "active_channels() --> number of currently active channels."},
  {"channels", PyAsteriskChannels, METH_VARARGS,
   "channels() --> return a list of active channel names."},
  {"hangup", PyAsteriskChannelHangup, METH_VARARGS,
   "hangup(name) --> hangup the named channel."},
  {"setvar", PyAsteriskChannelSetVar, METH_VARARGS,
   "setvar(channel, variable, value) --> set the given channel variable to value."},
  {"getvar", PyAsteriskChannelGetVar, METH_VARARGS,
   "getvar(channel, variable) --> get the value of the given channel variable."},
  {"chan_setattr", PyAsteriskChannelSetAttr, METH_VARARGS,
   "chan_setattr(channel, attr, value) -> set a channel state variable (attribute) to\n"
   "the given value."},
  {"chan_getattr", PyAsteriskChannelGetAttr, METH_VARARGS,
   "chan_getattr(channel, attribute) --> get a channel state variable (attribute)\n"},
  {"redirect", PyAsteriskRedirect, METH_VARARGS,
   "redirect(channel, (context, extension, priority) --> redirect the call to the given place."},
  {"originate_app", (PyCFunction)PyAsteriskOriginateApp, METH_VARARGS | METH_KEYWORDS,
   "originate_app((tech, data), (app, appdata), timeout=60, callerid=None, variables=None, accountcode='PyAsterisk')\n"
   "Originates a call using the specified tech and data, (i.e.: ('SIP', '123@example.com')) to the given app\n"
   "(i.e.: ('MeetMe', '000|mq')) with some optional arguments. Variables is of the form 'var=value|var=value|...'\n"
   "and causes channel variables to be set."},
  {"originate_exten", (PyCFunction)PyAsteriskOriginateExten, METH_VARARGS | METH_KEYWORDS,
   "originate_exten((tech, data), (context, exten, prio), timeout=60, callerid=None, variables=None, accountcode='PyAsterisk')\n"
   "Originates a call using the specified tech and data, (i.e.: ('SIP', '123@example.com')) to the given extension\n"
   "(i.e.: ('default', '1234', 1)) with some optional arguments. Variables is of the form 'var=value|var=value|...'\n"
   "and causes channel variables to be set."},
  {"contexts", PyAsteriskContexts, METH_VARARGS,
   "contexts() --> return a list of Asterisk dialplan contexts."},
  {"context_includes", PyAsteriskContextIncludes, METH_VARARGS,
   "context_includes(context) --> return a list of includes in the given context."},
  {"context_include_add", PyAsteriskContextIncludeAdd, METH_VARARGS,
   "context_include_add(context, include, registrar='PyAsterisk') -> add an include to context."},
  {"context_include_del", PyAsteriskContextIncludeDel, METH_VARARGS,
   "context_include_del(context, include, registrar='PyAsterisk') -> delete an include to context."},
  {"context_extensions", PyAsteriskContextExtensions, METH_VARARGS,
   "context_extensions(context) --> return a list ((exten, cid),...) of extensions present in the\n"
   "given context."},
  {"context_extension_get", PyAsteriskContextExtensionGet, METH_VARARGS,
   "context_extension_get(context, exten) --> [[prio, app, data, cid, registrar],...]\n"},
  {"context_extension_add", PyAsteriskContextExtensionAdd, METH_VARARGS,
   "context_extension_add(context, exten, prio, app, data="", replace=1, cid=None, registrar='PyAsterisk')\n"
   "add the given extension into the context"},
  {"context_extension_del", PyAsteriskContextExtensionDel, METH_VARARGS,
   "context_extension_del(context, exten, prio, cid, app, data, replace=1, registrar='PyAsterisk')\n"
   "delete the given extension from the context"},
  {NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC PyAsteriskInit(void)
{
	Py_InitModule("_asterisk", AsteriskMethods);
}

/*
 * Asterisk Module Housekeeping. Ver' simple.
 */
  
STANDARD_LOCAL_USER;
LOCAL_USER_DECL;

static void *python_init(void *dummy)
	/* ARGUSED */
{
  PyThreadState *tstate;

  ast_log(LOG_DEBUG, "Initializing Python Interpreter.\n");
  ast_verbose("Python %s %s\n", Py_GetPlatform(), Py_GetVersion());

  Py_SetProgramName(PyASTERISK_NAME);
  PyEval_InitThreads();
  Py_Initialize();
  PyAsteriskInit();

  /* Save the main thread... */
  tstate = PyEval_SaveThread();
  python_interp = tstate->interp;
  PyEval_RestoreThread(tstate);

  /* Run our intialization function, which
   * should not return 
   */
  PyRun_SimpleString("import sys\n"
		     "sys.path.append('" PyASTERISK_LIB_DIR "')\n"
		     "from PyAsteriskConfig import start\n"
		     "try:\n"
		     "    start()\n"
		     "except:\n"
		     "    from traceback import format_exception\n"
		     "    from sys import exc_info\n"
		     "    from string import join\n"
		     "    from _asterisk import verbose\n"
		     "    verbose(join(apply(format_exception, exc_info()), '\\n'))\n");

  ast_log(LOG_DEBUG, "Main Python thread returned");

  return NULL;
}

int unload_module(void)
{
  PyThreadState *tstate;

  STANDARD_HANGUP_LOCALUSERS;

  ast_log(LOG_DEBUG, "Terminating Python Interpreter.\n");

  if(python_interp) {
    /* interp is your reference to an interpreter object. */
    tstate = PyThreadState_New(python_interp);
    PyEval_AcquireThread(tstate);

    PyRun_SimpleString("from PyAsteriskConfig import stop\n"
		       "try:\n"
		       "    stop()\n"
		       "except:\n"
		       "    from traceback import format_exception\n"
		       "    from sys import exc_info\n"
		       "    from string import join\n"
		       "    from _asterisk import verbose\n"
		       "    verbose(join(apply(format_exception, exc_info()), '\\n'))\n");
  
    /* Release the thread. No Python API allowed beyond this point. */
    PyEval_ReleaseThread(tstate);
    PyThreadState_Delete(tstate);
  }

  Py_Finalize();
  python_interp = NULL;

  ast_log(LOG_DEBUG, "res_python.c unloaded.\n");
  return 0;
}

int load_module(void)
{
  ast_verbose(PyASTERISK_NAME " " PyASTERISK_VERSION "\n"
	      PyASTERISK_COPYRIGHT "\n");

  if(ast_pthread_create(&interp_thread, NULL, python_init, NULL)) {
    ast_log(LOG_ERROR, "Unable to create python interpreter\n");
    return -1;
  }
  ast_log(LOG_DEBUG, "res_python.c loaded.\n");
  return 0;
}

char *description(void)
{
        return tdesc;
}

int usecount(void)
{
        int res;
        STANDARD_USECOUNT(res);
        return res;
}

char *key()
{
        return ASTERISK_GPL_KEY;
}
